/**
 * SwissTM implementation.
 * 
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#include <iostream>

#include <cstdlib>

#include "common/version_lock.h"
#include "common/atomic.h"
#include "common/time.h"
#include "common/log.h"
#include "transaction_base.h"
#include "common/timestamp.h"
#include "constants.h"
#include "object.h"
#include "aborted_exc.h"

#ifdef USE_PREEMPTIVE_WAITING
#include "common/preemptive_utils.h"
#endif // USE_PREEMPTIVE_WAITING

// define logs to use for read and write sets
namespace swisstm {
	struct ReadSetEntry {
		ObjectHandleMixinvIRPrioT *handle;
		VersionLock version;
	};

	struct WriteSetEntry {
		ObjectHandleMixinvIRPrioT *handle;
		VersionLock version;
	};

	typedef Log<ReadSetEntry> ObjectHandleReadSet;
	typedef ObjectHandleReadSet::iterator ObjectHandleReadSetIterator;

	typedef Log<WriteSetEntry> ObjectHandleWriteSet;
	typedef ObjectHandleWriteSet::iterator ObjectHandleWriteSetIterator;	
}

namespace swisstm {

	class TxMixinvIRPrioT : public TxMixinvIRPrioTBase {
		public:
			TxMixinvIRPrioT();

			static void GlobalInit();

			void TxStart();
			void TxCommit();
			void TxAbort();
			void TxAbortFull();

			BaseObject *OpenWrite(ObjectHandleMixinvIRPrioT *handle);
			const BaseObject *OpenRead(ObjectHandleMixinvIRPrioT *handle);
			void Delete(ObjectHandleMixinvIRPrioT *handle);

		private:
			BaseObject *OpenWriteInner(ObjectHandleMixinvIRPrioT *handle);
			const BaseObject *OpenReadInner(ObjectHandleMixinvIRPrioT *handle);
			void DeleteInner(ObjectHandleMixinvIRPrioT *handle);

		private:
			bool Validate();
			bool Extend();

			// contention management
			bool ShouldAbortWrite(ObjectHandleMixinvIRPrioT *waitingFor);
			bool ShouldAbortDeleteInner(ObjectHandleMixinvIRPrioT *waitingFor);

			// cm policy
			bool CMStrongerThan(TxMixinvIRPrioT *other);

			void AbortSelf();

		private:
			ObjectHandleReadSet readSet;
			ObjectHandleWriteSet writeSet;

			const VersionLock locked_by_me;

			// contention management data
			volatile unsigned greedy_ts;
			volatile bool waiting;
			bool aborted_externally;

			unsigned succ_abort_cnt;

		private:
			void DecidePriority();

		public:
			// transaction state
			// TODO: think whether I need all these states
			enum TxState {
				TX_STATE_IDLE,
				TX_STATE_RUNNING,
				TX_STATE_STARTED_COMMITTING,
				TX_STATE_COMMIT_PREPARE,
				TX_STATE_COMMITTING
			};

			TxState state;

			// id
			const unsigned tid;
			const unsigned bitmask;

		private:
			class BoolWrapperSetFalse {
				public:
					BoolWrapperSetFalse(volatile bool *v) : val(v) {  }

					~BoolWrapperSetFalse() {
						*val = false;
					}

				private:
					volatile bool *val;
			};

			// use global array for allocation
			static TxMixinvIRPrioT descriptors[MAX_THREADS];

		public:
			void* operator new(size_t size) {
				return descriptors + get_tid();
			}

			void operator delete(void* ptr) {
				// do nothing
			}

		private:
			//util
			bool retrying() const {
				return succ_abort_cnt != 0;
			}

			bool first_retry() const {
				return succ_abort_cnt == 1;
			}

	} __attribute__ ((aligned(1024)));
}

//extern swisstm::GlobalTimestamp commit_ts;

extern swisstm::GlobalTimestamp cm_ts;

inline void swisstm::TxMixinvIRPrioT::GlobalInit() {
	// do nothing
}

inline swisstm::TxMixinvIRPrioT::TxMixinvIRPrioT() :
		locked_by_me(get_locked(this)),
		waiting(false), aborted_externally(false),
		succ_abort_cnt(0), state(TX_STATE_IDLE),
		tid(get_tid()), bitmask(BitsetBase::getBitmask(tid)) {
	register_transaction(tid, this);

	greedy_ts = 0;
}

inline void swisstm::TxMixinvIRPrioT::TxStart() {
	valid_ts = commit_ts.readCurrentTs();

	// get cm start time
	if(first_retry()) {
		greedy_ts = cm_ts.getNextTsAcquire() + 1;
	} else if(!retrying()){
		greedy_ts = MINIMUM_TS;
	}

	aborted_externally = false;
	state = TX_STATE_RUNNING;
}

inline void swisstm::TxMixinvIRPrioT::TxCommit() {
	if(!writeSet.empty()) {
		// announce start of the commit
		atomic_store_full(&state, TX_STATE_STARTED_COMMITTING);

		unsigned ts = commit_ts.getNextTsFull() + 1;

		// check read set consistency for non prioritized tx
		if(ts != valid_ts + 1 && !Validate()) {
			state = TX_STATE_IDLE;
			AbortSelf();
		}

		VersionLock commitVersion = get_version_lock(ts);

		// start committing values to written objects
		atomic_store_full(&state, TX_STATE_COMMITTING);

		VersionLock minVersion = get_version_lock(MINIMUM_TS);
		ObjectHandleWriteSetIterator writeIter;

		for(writeIter = writeSet.begin();writeIter.hasNext();writeIter.next()) {
			WriteSetEntry &entry = *writeIter;
			ObjectHandleMixinvIRPrioT *handle = entry.handle;
			atomic_store_release(&handle->old_version, minVersion);
			atomic_store_release(&handle->current, handle->log);

			atomic_store_release(&handle->old_version, commitVersion);
			atomic_store_release(&handle->lock, commitVersion);
		}

		// finish committing values
		atomic_store_release(&state, TX_STATE_IDLE);

		writeSet.clear();
	}

	readSet.clear();

	succ_abort_cnt = 0;
}

inline void swisstm::TxMixinvIRPrioT::TxAbort() {
	if(!writeSet.empty()) {
		ObjectHandleWriteSetIterator iter;

		for(iter = writeSet.begin();iter.hasNext();iter.next()) {
			WriteSetEntry &entry = *iter;
			entry.handle->lock = entry.version;
		}

		writeSet.clear();
	}

	readSet.clear();

	succ_abort_cnt++;

#ifdef USE_PREEMPTIVE_WAITING
	pre_yield();
#endif
}

inline void swisstm::TxMixinvIRPrioT::TxAbortFull() {
	TxAbort();

	succ_abort_cnt = 0;
}

inline const swisstm::BaseObject *swisstm::TxMixinvIRPrioT::OpenRead(
		ObjectHandleMixinvIRPrioT *handle) {
	BoolWrapperSetFalse wrapper(&waiting);
	return OpenReadInner(handle);
}

inline const swisstm::BaseObject *swisstm::TxMixinvIRPrioT::OpenReadInner(
		ObjectHandleMixinvIRPrioT *handle) {
	if(handle == NULL) {
		return NULL;
	}

	VersionLock version = handle->lock;

	// if locked by this tx return immediately
	if(version == locked_by_me) {
		return handle->log;
	}

	while(true) {
		TxState owner_state;
		VersionLock old_version;
		bool locked = is_locked(version);

		if(locked) {
			TxMixinvIRPrioT *owner = (TxMixinvIRPrioT *)get_owner(version);
			owner_state = (TxState)atomic_load_acquire(&owner->state);

			if(owner_state != TX_STATE_RUNNING &&
					owner_state != TX_STATE_IDLE) {
				// TODO: wait while it finishes here perhaps
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}

			old_version = (VersionLock)atomic_load_acquire(&handle->old_version);

			// ran into middle of object commit
			if(old_version == get_version_lock(MINIMUM_TS)) {
				version = (VersionLock)atomic_load_acquire(&handle->lock);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		BaseObject *ptr = (BaseObject *)atomic_load_acquire(&handle->current);

		if(locked) {
			VersionLock old_version_2 = (VersionLock)atomic_load_acquire(&handle->old_version);

			if(old_version != old_version_2) {
				version = (VersionLock)atomic_load_acquire(&handle->lock);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		} else { // TODO: think about this
			VersionLock version_2 = (VersionLock)atomic_load_acquire(&handle->lock);

			if(version != version_2) {
				version = version_2;
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		VersionLock ver = locked ? old_version : version;

		// check validity
		if(get_value(ver) > valid_ts) {
			if(!Extend()) {
				// must abort here
				AbortSelf();
			}
		}

		// here all is consistent and valid :)
		ReadSetEntry *entry = readSet.get_next();
		entry->handle = handle;
		entry->version = ver;
		return ptr;
	}
}

// TODO: think about this
inline bool swisstm::TxMixinvIRPrioT::Validate() {
	ObjectHandleReadSetIterator readSetIter;

	for(readSetIter = readSet.begin();readSetIter.hasNext();readSetIter.next()) {
		ReadSetEntry &entry = *readSetIter;

		VersionLock currentVersion = (VersionLock)atomic_load_acquire(
			&entry.handle->lock);

		if(is_locked(currentVersion)) {
			if(currentVersion != locked_by_me) {
				TxMixinvIRPrioT *owner = (TxMixinvIRPrioT *)get_owner(currentVersion);
				TxState owner_state = (TxState)atomic_load_acquire(&owner->state);

				if(owner_state != TX_STATE_RUNNING &&
						owner_state != TX_STATE_IDLE) {
					return false;
				}

				VersionLock old_version = (VersionLock)atomic_load_acquire(
					&entry.handle->old_version);

				if(old_version != entry.version) {
					return false;
				}
			}
		} else if(currentVersion != entry.version) {
			return false;
		}
	}

	return true;
}

inline bool swisstm::TxMixinvIRPrioT::Extend() {
	unsigned ts = commit_ts.readCurrentTs();

	if(Validate()) {
		valid_ts = ts;
		return true;
	}

	return false;
}

inline swisstm::BaseObject *swisstm::TxMixinvIRPrioT::OpenWrite(
			ObjectHandleMixinvIRPrioT *handle) {
	BoolWrapperSetFalse wrapper(&waiting);
	return OpenWriteInner(handle);
}

inline swisstm::BaseObject *swisstm::TxMixinvIRPrioT::OpenWriteInner(
		ObjectHandleMixinvIRPrioT *handle) {
	if(handle == NULL) {
		return NULL;
	}

	VersionLock version = handle->lock;

	// if is locked by this tx return immediately
	if(version == locked_by_me) {
		return handle->log;
	}

	while(true) {
		if(is_locked(version)) {
			if(ShouldAbortWrite(handle)) {
				AbortSelf();
			} else {
				version = (VersionLock)atomic_load_no_barrier(&handle->lock);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		// now check if the version is ok - this must be done as the object
		// might be used later for reading
		// not neccessary if tx is prioritized
		if(get_value(version) > valid_ts) {
			if(!Extend()) {
				// must abort here
				AbortSelf();
			}
		}

		if(atomic_cas_acquire(&handle->lock, version, locked_by_me)) {
			// now the object is locked and the snapshot is still valid
			BaseObject *current = handle->current;
			BaseObject *ret = current->cloneFull();
			tx_free(current);
			handle->log = ret;

			WriteSetEntry *entry = writeSet.get_next();
			entry->handle = handle;
			entry->version = version;

			return ret;
		} else {
			version = (VersionLock)atomic_load_no_barrier(&handle->lock);
#ifdef USE_PREEMPTIVE_WAITING
			pre_yield();
#endif
		}
	}

	// this should never happen
	return NULL;
}

inline void swisstm::TxMixinvIRPrioT::Delete(ObjectHandleMixinvIRPrioT *handle) {
	BoolWrapperSetFalse wrapper(&waiting);
	return DeleteInner(handle);
}

inline void swisstm::TxMixinvIRPrioT::DeleteInner(ObjectHandleMixinvIRPrioT *handle) {
	if(handle == NULL) {
		return;
	}

	// if was locked by me DeleteInner immediately
	VersionLock version = handle->lock;

	if(version == locked_by_me) {
		tx_free(handle->log);
		handle->log = NULL;
		freeHandle(handle);
		return;
	}

	while(true) {
		if(is_locked(version)) {
			if(ShouldAbortDeleteInner(handle)) {
				AbortSelf();
			} else {
				version = (VersionLock)atomic_load_no_barrier(&handle->lock);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		if(get_value(version) > valid_ts) {
			if(!Extend()) {
				// must abort here
				AbortSelf();
			}
		}

		// now try to cas it into the handle
		if(atomic_cas_acquire(&handle->lock, version, locked_by_me)) {
			handle->log = NULL;
			tx_free(handle->current);
			freeHandle(handle);

			WriteSetEntry *entry = writeSet.get_next();
			entry->handle = handle;
			entry->version = version;

			return;
		} else {
			version = (VersionLock)atomic_load_no_barrier(&handle->lock);
#ifdef USE_PREEMPTIVE_WAITING
			pre_yield();
#endif
		}
	}
}

inline bool swisstm::TxMixinvIRPrioT::ShouldAbortWrite(
		ObjectHandleMixinvIRPrioT *waitingFor) {
	if(aborted_externally) {
		return true;
	}

	if(greedy_ts == 0) {
		return true;
	}

	VersionLock version = waitingFor->lock;

	if(is_locked(version)) {
		TxMixinvIRPrioT *owner = (TxMixinvIRPrioT *)get_owner(version);

		if(CMStrongerThan(owner)) {
			owner->aborted_externally = true;
			return false;
		}

		return true;

		waiting = true;
	}

	return false;
}

inline bool swisstm::TxMixinvIRPrioT::ShouldAbortDeleteInner(
		ObjectHandleMixinvIRPrioT *waitingFor) {
	return ShouldAbortWrite(waitingFor);
}

inline bool swisstm::TxMixinvIRPrioT::CMStrongerThan(
		TxMixinvIRPrioT *other) {
	unsigned other_ts = other->greedy_ts;

	if(greedy_ts == MINIMUM_TS) {
		return false;
	}

	return other_ts == MINIMUM_TS ||
			(greedy_ts < other_ts || other->waiting);
}

inline void swisstm::TxMixinvIRPrioT::AbortSelf() {
	throw AbortedException();
}
